/******************************************************************************/
/* PFile implementation file.                                                 */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#define _BUILDING_POLYKIT_LIBRARY_

// PolyKit headers
#include "PDebug.h"
#include "PException.h"
#include "PFile.h"
#include "PDirectory.h"
#include "GlobalFuncs.h"


/******************************************************************************/
/* PFile class                                                                */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/******************************************************************************/
PFile::PFile(void)
{
	fileOpened   = false;
	eof          = false;

#if __p_os == __p_beos
	file = NULL;

#elif __p_os == __p_windows
	file = INVALID_HANDLE_VALUE;

#elif __p_os == __p_linux
implement if needed
#endif
}



/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "fileName" is the name of the file to open.                        */
/*         "openFlags" is the open flags you want the file to be opened with. */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PFile::PFile(PString fileName, uint16 openFlags)
{
	fileOpened = false;
	eof        = false;

	Open(fileName, openFlags);
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PFile::~PFile(void)
{
	Close();
}



/******************************************************************************/
/* Open() will open the file with the open flags given.                       */
/*                                                                            */
/* Input:  "openFlags" is the open flags you want the file to be opened with. */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Open(uint16 openFlags)
{
	Open(name, openFlags);
}



/******************************************************************************/
/* Open() will open the file with the name and open flags given.              */
/*                                                                            */
/* Input:  "fileName" is the name of the file to open.                        */
/*         "openFlags" is the open flags you want the file to be opened with. */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Open(PString fileName, uint16 openFlags)
{
	ASSERT(fileOpened == false);

	name = fileName;

#if __p_os == __p_beos

	// Find out the open flags
	uint32 openMode;
	status_t error;

	switch (openFlags & 0x0007)
	{
		case pModeRead:
			openMode = B_READ_ONLY;
			break;

		case pModeWrite:
			openMode = B_WRITE_ONLY;
			break;

		case pModeReadWrite:
			openMode = B_READ_WRITE;
			break;

		default:
			// Open mode not supported
			ASSERT(false);
			openMode = B_READ_ONLY;
			break;
	}

	// Or extra flags if any
	if (openFlags & pModeCreate)
		openMode |= (B_CREATE_FILE | B_ERASE_FILE);

	if (openFlags & pModeNoTruncate)
		openMode &= ~(B_ERASE_FILE);

	if (openFlags & pModeFailIfExists)
		openMode |= B_FAIL_IF_EXISTS;

	// Now open the file
	file = new BFile();

	error = file->SetTo(fileName, openMode);
	if (error != B_OK)
	{
		delete file;
		throw PFileException(ConvertOSError(error), fileName);
	}

#elif __p_os == __p_windows

	// Handle the open flags
	DWORD desiredAccess;
	DWORD shareMode;
	DWORD creationDisposition;

	// Store the open flags concerning desired access
	prevOpenFlags = openFlags & 0x0007;

	// Handle desired access
	switch (prevOpenFlags)
	{
		// Read only mode
		case pModeRead:
			desiredAccess = GENERIC_READ;
			shareMode = FILE_SHARE_READ;	// Subsequent open operations on the file is allowed
			break;

		// Write only mode
		case pModeWrite:
			desiredAccess = GENERIC_WRITE;
			shareMode = 0;					// The file cannot be shared
			break;

		// Read-write mode
		case pModeReadWrite:
			desiredAccess = GENERIC_READ | GENERIC_WRITE;
			shareMode = 0;					// The file cannot be shared
			break;

		// Default mode is read only mode
		default:
			// Open mode not supported
			ASSERT(false);
			desiredAccess = GENERIC_READ;
			shareMode = FILE_SHARE_READ;	// Subsequent open operations on the file is allowed
			break;
	}

	// Handle creation disposition

	// Ff the desired access mode is read only, then open the file as an existing file
	if (desiredAccess == GENERIC_READ)
		creationDisposition = OPEN_EXISTING;

	// Else the desired access mode is read-write or write only
	else
	{
		// Check if the a file should be created if it does not exist already
		if (openFlags & pModeCreate)
		{
			// Check if creating the file should fail if it does not exist already.

			// If so, create the file as a new file.
			if (openFlags & pModeFailIfExists)
				creationDisposition = CREATE_NEW;

			// Else the file should always overwrite an existing one
			else
			{
				// Check if no truncation should be done on the file

				// If so, just open the file if it exist or create a new file otherwise
				if (openFlags & pModeNoTruncate)
					creationDisposition = OPEN_ALWAYS;

				// Else the file should be created by overwriting an existing file
				// or create a new file if it does not exists.
				else
					creationDisposition = CREATE_ALWAYS;
			}
		}
		// If not, assume the file already exist
		else
		{
			// If no truncating should be done, then just open the file as an existing one
			if (openFlags & pModeNoTruncate)
				creationDisposition = OPEN_EXISTING;

			// Else truncate an existing file
			else
				creationDisposition = TRUNCATE_EXISTING;
		}
	}

	// Attempt to create the file
	file = CreateFile(fileName, desiredAccess, shareMode, NULL, creationDisposition, FILE_ATTRIBUTE_NORMAL, NULL);
	if (file == INVALID_HANDLE_VALUE)
		throw PFileException(ConvertOSError(GetLastError()), name);	

#elif __p_os == __p_linux
implement
#endif

	// Set the opened file flag
	fileOpened = true;
}



/******************************************************************************/
/* Close() closes the file again.                                             */
/******************************************************************************/
void PFile::Close(void)
{
	if (fileOpened)
	{
#if __p_os == __p_beos

		delete file;
		file = NULL;

#elif __p_os == __p_windows

		if (file != INVALID_HANDLE_VALUE)
			VERIFY(CloseHandle(file));
		file = INVALID_HANDLE_VALUE;

#elif __p_os == __p_linux
implement
#endif
	}

	// Cleanup class variables
	fileOpened = false;
}



/******************************************************************************/
/* IsOpen() checks to see if the file is opened.                              */
/******************************************************************************/
bool PFile::IsOpen(void) const
{
	return (fileOpened);
}



/******************************************************************************/
/* Read() will read from the file and store the data into the buffer given.   */
/*                                                                            */
/* Input:  "buffer" is a pointer to the buffer to store the data in.          */
/*         "count" is the number of bytes you want to read.                   */
/*                                                                            */
/* Output: Is the actual number of bytes read.                                */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PFile::Read(void *buffer, int32 count)
{
	int32 readBytes;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	readBytes = file->Read(buffer, count);
	if (readBytes < 0)
		throw PFileException(ConvertOSError(readBytes), name);

#elif __p_os == __p_windows

	// Read the file
	if (!ReadFile(file, buffer, count, (unsigned long *)&readBytes, NULL))
		throw PFileException(ConvertOSError(GetLastError()), name);

#elif __p_os == __p_linux
implement
#endif

	// Set the EOF flag
	eof = (readBytes == 0) ? true : false;

	return (readBytes);
}



/******************************************************************************/
/* Write() will write into the file from the buffer given.                    */
/*                                                                            */
/* Input:  "buffer" is a pointer to the buffer where the data to write is.    */
/*         "count" is the number of bytes you want to write.                  */
/*                                                                            */
/* Output: Is the actual number of bytes written.                             */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PFile::Write(const void *buffer, int32 count)
{
	int32 writtenBytes;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	writtenBytes = file->Write(buffer, count);
	if (writtenBytes < 0)
		throw PFileException(ConvertOSError(writtenBytes), name);

#elif __p_os == __p_windows

	if (!WriteFile(file, buffer, count, (unsigned long *)&writtenBytes, NULL))
	{
		DWORD error = GetLastError();

		ASSERT(false);

		// Throw exception?
	}

#elif __p_os == __p_linux
implement
#endif

	return (writtenBytes);
}



/******************************************************************************/
/* ReadString() reads a string into the buffer.                               */
/*                                                                            */
/* Input:  "buffer" is a pointer a buffer to store the string in.             */
/*         "number" is the max number to read.                                */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::ReadString(char *buffer, int32 number)
{
	uint32 bytesRead;

	bytesRead = Read(buffer, number);
	buffer[bytesRead] = 0x00;
}



/******************************************************************************/
/* ReadString() reads a string in Pascal format and return it.                */
/*                                                                            */
/* Output: The read string.                                                   */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PFile::ReadString(void)
{
	PString str;
	uint16 len;
	char *buffer;

	// Read the length of the string
	len = Read_L_UINT16();

	if (len != 0)
	{
		buffer = str.GetBufferSetLength(len);
		Read(buffer, len);
		str.ReleaseBuffer(len);
	}

	return (str);
}



/******************************************************************************/
/* Read_UINT8() reads one byte and return it.                                 */
/*                                                                            */
/* Output: The read byte.                                                     */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint8 PFile::Read_UINT8(void)
{
	uint8 retVal = 0;

	// Read the byte
	Read(&retVal, 1);

	return (retVal);
}



/******************************************************************************/
/* Read_L_UINT16() reads 2 bytes in little endian and return it in host       */
/*      format.                                                               */
/*                                                                            */
/* Output: The read 16 bit number.                                            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint16 PFile::Read_L_UINT16(void)
{
	uint16 retVal = 0;

	// Read the number
	Read(&retVal, 2);

	return (P_LENDIAN_TO_HOST_INT16(retVal));
}



/******************************************************************************/
/* Read_L_UINT32() reads 4 bytes in little endian and return it in host       */
/*      format.                                                               */
/*                                                                            */
/* Output: The read 32 bit number.                                            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint32 PFile::Read_L_UINT32(void)
{
	uint32 retVal = 0;

	// Read the number
	Read(&retVal, 4);

	return (P_LENDIAN_TO_HOST_INT32(retVal));
}



/******************************************************************************/
/* ReadArray_L_UINT16s() reads an array of uint16 in little endian and        */
/*      convert them to host format.                                          */
/*                                                                            */
/* Input:  "buffer" is a pointer to where to store the data.                  */
/*         "number" is the number of elements to read.                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::ReadArray_L_UINT16s(uint16 *buffer, int32 number)
{
	for (int32 t = 0; t < number; t++)
		*buffer++ = Read_L_UINT16();
}



/******************************************************************************/
/* ReadArray_L_UINT32s() reads an array of uint32 in little endian and        */
/*      convert them to host format.                                          */
/*                                                                            */
/* Input:  "buffer" is a pointer to where to store the data.                  */
/*         "number" is the number of elements to read.                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::ReadArray_L_UINT32s(uint32 *buffer, int32 number)
{
	for (int32 t = 0; t < number; t++)
		*buffer++ = Read_L_UINT32();
}



/******************************************************************************/
/* Read_B_UINT16() reads 2 bytes in big endian and return it in host format.  */
/*                                                                            */
/* Output: The read 16 bit number.                                            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint16 PFile::Read_B_UINT16(void)
{
	uint16 retVal = 0;

	// Read the number
	Read(&retVal, 2);

	return (P_BENDIAN_TO_HOST_INT16(retVal));
}



/******************************************************************************/
/* Read_B_UINT32() reads 4 bytes in big endian and return it in host format.  */
/*                                                                            */
/* Output: The read 32 bit number.                                            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint32 PFile::Read_B_UINT32(void)
{
	uint32 retVal = 0;

	// Read the number
	Read(&retVal, 4);

	return (P_BENDIAN_TO_HOST_INT32(retVal));
}



/******************************************************************************/
/* ReadArray_B_UINT16s() reads an array of uint16 in big endian and convert   */
/*      them to host format.                                                  */
/*                                                                            */
/* Input:  "buffer" is a pointer to where to store the data.                  */
/*         "number" is the number of elements to read.                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::ReadArray_B_UINT16s(uint16 *buffer, int32 number)
{
	for (int32 t = 0; t < number; t++)
		*buffer++ = Read_B_UINT16();
}



/******************************************************************************/
/* ReadArray_B_UINT32s() reads an array of uint32 in big endian and convert   */
/*      them to host format.                                                  */
/*                                                                            */
/* Input:  "buffer" is a pointer to where to store the data.                  */
/*         "number" is the number of elements to read.                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::ReadArray_B_UINT32s(uint32 *buffer, int32 number)
{
	for (int32 t = 0; t < number; t++)
		*buffer++ = Read_B_UINT32();
}



/******************************************************************************/
/* WriteString() writes a string in Pascal format into the file.              */
/*                                                                            */
/* Input:  "string" is the string to write.                                   */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::WriteString(PString string)
{
	uint32 len;
	char *str;

	len = string.GetLength();
	Write_L_UINT16((uint16)len);

	if (len != 0)
	{
		str = string.GetBuffer(1);
		Write(str, len);
	}
}



/******************************************************************************/
/* Write_UINT8() writes one byte.                                             */
/*                                                                            */
/* Input:  "num" is the byte to write.                                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Write_UINT8(uint8 num)
{
	Write(&num, 1);
}



/******************************************************************************/
/* Write_L_UINT16() writes 2 bytes in little endian.                          */
/*                                                                            */
/* Input:  "num" is the 16 bit number to write.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Write_L_UINT16(uint16 num)
{
	uint16 newNum;

	newNum = P_HOST_TO_LENDIAN_INT16(num);
	Write(&newNum, 2);
}



/******************************************************************************/
/* Write_L_UINT32() writes 4 bytes in little endian.                          */
/*                                                                            */
/* Input:  "num" is the 32 bit number to write.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Write_L_UINT32(uint32 num)
{
	uint32 newNum;

	newNum = P_HOST_TO_LENDIAN_INT32(num);
	Write(&newNum, 4);
}



/******************************************************************************/
/* WriteArray_L_UINT16s() writes an array of uint16 in little endian.         */
/*                                                                            */
/* Input:  "buffer" is a pointer to the data to write.                        */
/*         "number" is the number of elements to write.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::WriteArray_L_UINT16s(const uint16 *buffer, int32 number)
{
	for (int32 t = 0; t < number; t++)
		Write_L_UINT16(*buffer++);
}



/******************************************************************************/
/* WriteArray_L_UINT32s() writes an array of uint32 in little endian.         */
/*                                                                            */
/* Input:  "buffer" is a pointer to the data to write.                        */
/*         "number" is the number of elements to write.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::WriteArray_L_UINT32s(const uint32 *buffer, int32 number)
{
	for (int32 t = 0; t < number; t++)
		Write_L_UINT32(*buffer++);
}



/******************************************************************************/
/* Write_B_UINT16() writes 2 bytes in big endian.                             */
/*                                                                            */
/* Input:  "num" is the 16 bit number to write.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Write_B_UINT16(uint16 num)
{
	uint16 newNum;

	newNum = P_HOST_TO_BENDIAN_INT16(num);
	Write(&newNum, 2);
}



/******************************************************************************/
/* Write_B_UINT32() writes 4 bytes in big endian.                             */
/*                                                                            */
/* Input:  "num" is the 32 bit number to write.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Write_B_UINT32(uint32 num)
{
	uint32 newNum;

	newNum = P_HOST_TO_BENDIAN_INT32(num);
	Write(&newNum, 4);
}



/******************************************************************************/
/* WriteArray_B_UINT16s() writes an array of uint16 in big endian.            */
/*                                                                            */
/* Input:  "buffer" is a pointer to the data to write.                        */
/*         "number" is the number of elements to write.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::WriteArray_B_UINT16s(const uint16 *buffer, int32 number)
{
	for (int32 t = 0; t < number; t++)
		Write_B_UINT16(*buffer++);
}



/******************************************************************************/
/* WriteArray_B_UINT32s() writes an array of uint32 in big endian.            */
/*                                                                            */
/* Input:  "buffer" is a pointer to the data to write.                        */
/*         "number" is the number of elements to write.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::WriteArray_B_UINT32s(const uint32 *buffer, int32 number)
{
	for (int32 t = 0; t < number; t++)
		Write_B_UINT32(*buffer++);
}



/******************************************************************************/
/* Seek() will seek to a new position in the file.                            */
/*                                                                            */
/* Input:  "offset" is the number of bytes you want to move.                  */
/*         "from" is where you want to move from.                             */
/*                                                                            */
/* Output: Is the new position.                                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PFile::Seek(int32 offset, PSeekFlags from)
{
	int32 newPos;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	int32 seekMode;

	// Convert the from flag to a seek mode flag
	switch (from)
	{
		case pSeekBegin:
			seekMode = SEEK_SET;
			break;

		case pSeekCurrent:
			seekMode = SEEK_CUR;
			break;

		case pSeekEnd:
			seekMode = SEEK_END;
			break;

		default:
			// Unknown seek flag
			ASSERT(false);
			seekMode = SEEK_SET;
			break;
	}

	// Seek the file position
	newPos = file->Seek(offset, seekMode);
	if (newPos < 0)
		throw PFileException(ConvertOSError(newPos), name);

#elif __p_os == __p_windows

	DWORD moveMethod;

	// Handle seek flags
	switch (from)
	{
		case pSeekBegin:
			moveMethod = FILE_BEGIN;
			break;

		case pSeekCurrent:
			moveMethod = FILE_CURRENT;
			break;

		case pSeekEnd:
			moveMethod = FILE_END;
			break;

		default:
			// Unknown seek flag
			ASSERT(false);
			moveMethod = FILE_BEGIN;
			break;
	}

	// Set the file position
	newPos = SetFilePointer(file, offset, NULL, moveMethod);
	if (newPos == INVALID_SET_FILE_POINTER)
		throw PFileException(ConvertOSError(GetLastError()), name);

#elif __p_os == __p_linux
implement
#endif

	// Clear eof flag
	eof = false;

	return (newPos);
}



/******************************************************************************/
/* SeekToBegin() will seek to the beginning of the file.                      */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::SeekToBegin(void)
{
	Seek(0, PFile::pSeekBegin);
}



/******************************************************************************/
/* SeekToEnd() will seek to the end of the file.                              */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::SeekToEnd(void)
{
	Seek(0, PFile::pSeekEnd);
}



/******************************************************************************/
/* GetPosition() will return the current file position.                       */
/*                                                                            */
/* Output: Is the current file position.                                      */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PFile::GetPosition(void) const
{
	int32 pos;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	pos = file->Position();
	if (pos < 0)
		throw PFileException(ConvertOSError(pos), name);

#elif __p_os == __p_windows

	// Set the file position
	pos = SetFilePointer(file, 0, NULL, FILE_CURRENT);
	if (pos == INVALID_SET_FILE_POINTER)
		throw PFileException(ConvertOSError(GetLastError()), name);

#elif __p_os == __p_linux
implement
#endif

	return (pos);
}



/******************************************************************************/
/* IsEOF() will find out if the file pointer is beyond the size of the file.  */
/*                                                                            */
/* Output: Is true if End Of File has been reached, else false.               */
/******************************************************************************/
bool PFile::IsEOF(void) const
{
	return (eof);
}



/******************************************************************************/
/* GetLength() will return the length of the file.                            */
/*                                                                            */
/* Output: Is the length of the file.                                         */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PFile::GetLength(void) const
{
	int32 length;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	status_t error;
	int64 realLength;

	error = file->GetSize(&realLength);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

	length = realLength;

#elif __p_os == __p_windows

	// Get the file size
	length = GetFileSize(file, NULL);
	if (length == -1)
		throw PFileException(ConvertOSError(GetLastError()), name);

#elif __p_os == __p_linux
implements
#endif

	return (length);
}



/******************************************************************************/
/* SetLength() will change the length of the file.                            */
/*                                                                            */
/* Input:  "newLength" is the new length of the file.                         */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::SetLength(int32 newLength)
{
	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	status_t error;

	error = file->SetSize(newLength);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

#elif __p_os == __p_windows

	// Get the current position
	int32 oldPos = GetPosition();

	// Seek the position equal to the new length of the file
	Seek(newLength, PFile::pSeekBegin);

	// Move the end-of-file (EOF) position
	if (!SetEndOfFile(file))
		throw PFileException(ConvertOSError(GetLastError()), name);

	// Restore the old position
	Seek(oldPos, PFile::pSeekBegin);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* GetFileName() will return the filename of the current file.                */
/*                                                                            */
/* Output: Is the filename without the path.                                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PFile::GetFileName(void) const
{
	PString fileName;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	BEntry entry;
	char nameBuf[B_FILE_NAME_LENGTH];
	status_t error;

	// Set the entry to the file
	error = entry.SetTo(name);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

	// Get the name of the entry
	error = entry.GetName(nameBuf);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

	// Return the name
	fileName = nameBuf;

#elif __p_os == __p_windows
	
	// Use PDirectory to return file name of the path
	PDirectory dir;
	fileName = dir.GetFilePart(name);

	// The file name must not be empty
	ASSERT(!name.IsEmpty());

#elif __p_os == __p_linux
implement
#endif

	return (fileName);
}



/******************************************************************************/
/* GetFilePath() will return the expanded path without the filename of the    */
/*      current file. See the GetFullPath() for more information.             */
/*                                                                            */
/* Output: Is the path without the filename.                                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PFile::GetFilePath(void) const
{
	PString path;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	BPath filePath;
	PString fullPath, fileName;
	status_t error;

	// Set the entry to the file
	error = filePath.SetTo(name);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

	// Get the path
	fullPath = filePath.Path();
	fileName = filePath.Leaf();

	// Return the path
	path = fullPath.Left(fullPath.GetByteLength() - fileName.GetByteLength());

#elif __p_os == __p_windows

	// Get the full path name
	path = GetFullPath();

	// Remove the file name from the path name
	path = path.Left(path.ReverseFind(P_DIRSLASH_C) + 1);

#elif __p_os == __p_linux
implement
#endif

	return (path);
}



/******************************************************************************/
/* GetFullPath() will return the path with the filename of the current file.  */
/*      It will expand the path if the file is opened with a relative path,   */
/*      e.g. if the current directory is: /boot/home and the relative path is */
/*      ../MyDir/MyDir2/../MyFile.txt   ->   /boot/MyDir/MyFile.txt           */
/*                                                                            */
/* Output: Is the full path with the filename.                                */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PFile::GetFullPath(void) const
{
	PString path;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	BPath filePath;
	status_t error;

	// Set the entry to the file
	error = filePath.SetTo(name);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

	// Get the path
	path = filePath.Path();

#elif __p_os == __p_windows

	char pathBuffer[MAX_PATH];
	char *filePath;

	// Get the full path name
	if (GetFullPathName(name, MAX_PATH, pathBuffer, &filePath) == 0)
		throw PFileException(ConvertOSError(GetLastError()), name);

	path = pathBuffer;

#elif __p_os == __p_linux
implement
#endif

	return (path);
}



/******************************************************************************/
/* GetFileType() will return the type of the file.                            */
/*      BeOS   : The mimestring.                                              */
/*      Windows: The file extension.                                          */
/*      Linux  : ???                                                          */
/*                                                                            */
/* Output: Is the file type.                                                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PFile::GetFileType(void) const
{
	PString type;

	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	BNodeInfo nodeInfo;
	char buffer[B_MIME_TYPE_LENGTH];
	status_t error;

	// Set the node info to the node
	error = nodeInfo.SetTo(file);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

	// Get the mime string
	error = nodeInfo.GetType(buffer);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

	// Return the type
	type = buffer;

#elif __p_os == __p_windows
	
	int32 pos;

	// Find the last dot, that is the beginning of the file extension
	pos = name.ReverseFind('.');

	// Return an empty string if no extension could be found
	if (pos == -1)
		return ("");

	// Assign the type string to the extension string
	type = name.Mid(pos + 1);

#elif __p_os == __p_linux
implement
#endif

	return (type);
}



/******************************************************************************/
/* SetFilePath() will set the class to use the new filename. It does not open */
/*      the file, it just associates the object with the name.                */
/*                                                                            */
/* Input:  "newName" is the new name to use.                                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::SetFilePath(PString newName)
{
	ASSERT(fileOpened == false);

	name = newName;
}



/******************************************************************************/
/* SetFileType() will change the type of the file.                            */
/*      BeOS   : Changes the mimestring.                                      */
/*      Windows: Renames the file extension.                                  */
/*      Linux  : ???                                                          */
/*                                                                            */
/* Input:  "newType" is a string with the new file type.                      */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::SetFileType(PString newType)
{
	ASSERT(fileOpened == true);

#if __p_os == __p_beos

	BNodeInfo nodeInfo;
	status_t error;

	// Set the node info to the node
	error = nodeInfo.SetTo(file);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

	// Change the mime string
	error = nodeInfo.SetType(newType);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), name);

#elif __p_os == __p_windows

	// The file must be opened with write access
	if ((prevOpenFlags != pModeWrite) && (prevOpenFlags != pModeReadWrite))
		throw PFileException(P_FILE_ERR_BAD_ACCESS_MODE);	// Bad access mode

	// Store current file position
	int32 curPos = GetPosition();

	// Cut away the old file extension from the file name
	PString newName = name.Left(name.ReverseFind('.') + 1);

	// Add the new file extension
	newName += newType;

	// Close the current file before renaming to avoid access violation
	Close();

	// Try to rename the file
	try
	{
		Rename(name, newName);
	}
	// Catch all exceptions
	catch(...)
	{
		// Cleanup
		try
		{
			// Open the file using the new name
			Open(newName, prevOpenFlags);

			// Set the current file position
			VERIFY(Seek(curPos, pSeekBegin) == curPos);
		}

		// Catch all exception
		catch(...)
		{
			// Play beep to alert the user of a fatal operation
			PlayBeep();
		}

		// Throw the exception
		throw;
	}

	// Open the file using the new name
	Open(newName, prevOpenFlags);

	// Set the current file position
	VERIFY(Seek(curPos, pSeekBegin) == curPos);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Rename() will rename a file.                                               */
/*                                                                            */
/* Input:  "oldName" is the name of the file you want to rename.              */
/*         "newName" is the name you want to rename to.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Rename(PString oldName, PString newName)
{
#if __p_os == __p_beos

	BEntry entry;
	status_t error;

	// Initialize the entry with the old name
	error = entry.SetTo(oldName);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), oldName);

	// Now rename the file
	error = entry.Rename(newName);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), oldName);

#elif __p_os == __p_windows

	if (newName.ReverseFind(P_DIRSLASH_C) == -1)
	{
		PDirectory dir;
		newName = dir.GetDirectoryPart(oldName) + newName;
	}

	// Rename the file by moving it
	if (!MoveFile(oldName, newName))
		throw PFileException(ConvertOSError(GetLastError()), oldName);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Remove() will remove a file.                                               */
/*                                                                            */
/* Input:  "fileName" is the name of the file you want to remove.             */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::Remove(PString fileName)
{
#if __p_os == __p_beos

	BEntry entry;
	status_t error;

	// Initialize the entry with the old name
	error = entry.SetTo(fileName);
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), fileName);

	// Now rename the file
	error = entry.Remove();
	if (error != B_OK)
		throw PFileException(ConvertOSError(error), fileName);

#elif __p_os == __p_windows

	// Delete the file
	if (!DeleteFile(fileName))
		throw PFileException(ConvertOSError(GetLastError()), fileName);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* FileExists() will test to see if the file exists.                          */
/*                                                                            */
/* Input:  "fileName" is the name of the file you want to check.              */
/*                                                                            */
/* Output: True if the file exists, else false.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
bool PFile::FileExists(PString fileName)
{
#if __p_os == __p_beos

	BFile file;
	status_t error;

	// Try to open the file
	error = file.SetTo(fileName, B_READ_ONLY);
	if (error == B_OK)
		return (true);

	return (false);

#elif __p_os == __p_windows

	HANDLE file;
	WIN32_FIND_DATA findData;

	// Find the file
	file = FindFirstFile(fileName, &findData);

	// Check if an error occured
	if (file == INVALID_HANDLE_VALUE)
	{
		// Get the last error
		DWORD error = GetLastError();

		// Check if the file path was not found
		if (error == ERROR_PATH_NOT_FOUND)
			return (false);						// If so, the file do not exist

		// Check for sharing violation
		if (error == ERROR_SHARING_VIOLATION)
			return (true);						// If so, the file exist

		// Throw a file exception
		throw PFileException(ConvertOSError(error), fileName);
	}

	// Close the file
	FindClose(file);

	// The file was found
	return (true);

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* ReadLine() reads an ASCII line from the file and return it. It will filter */
/*      out the new line codes.                                               */
/*                                                                            */
/* Output: The line read.                                                     */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PFile::ReadLine(void)
{
	PString line;
	char *strBuf, *workBuf;
	int32 readBytes, i;

	// Allocate space to hold the string
	strBuf = line.GetBufferSetLength(512);

	// Fill the buffer
	readBytes = Read(strBuf, 512);
	workBuf   = strBuf;

	if (readBytes > 0)
	{
		// Find the end of the string
		for (i = 0; i < readBytes; i++, workBuf++)
		{
			if ((*workBuf == 0x0a) || (*workBuf == 0x00))
				break;
		}

		if (i == 512)
			throw PFileException(P_FILE_ERR_EOF, name);

		// Seek back in the file
		Seek(-(readBytes - i - 1), pSeekCurrent);

		// Check for 0x0d, 0x0a format (Windows)
		if (workBuf != strBuf)
			if (*(workBuf - 1) == 0x0d)
				workBuf--;
	}

	// Null terminate the string and clean up
	*workBuf = 0x00;
	line.ReleaseBuffer();

	return (line);
}



/******************************************************************************/
/* WriteLine() writes an ASCII line to the file including the new line code.  */
/*                                                                            */
/* Input:  "line" is the string to write.                                     */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PFile::WriteLine(PString line)
{
#if (__p_os == __p_beos) || (__p_os == __p_linux)

	char cr = 0x0a;

#elif __p_os == __p_windows

	char cr[2] = {0x0d, 0x0a};

#endif

	char *str;

	// Write the line
	if (line.GetLength() != 0)
	{
		str = line.GetBuffer(1);
		Write(str, line.GetLength());
	}

#if (__p_os == __p_beos) || (__p_os == __p_linux)

	// Add the new line code
	Write(&cr, 1);

#elif __p_os == __p_windows

	// Add the new line code
	Write(&cr, 2);

#endif
}





/******************************************************************************/
/* PCacheFile class                                                           */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "size" is the size of the cache in bytes.                          */
/******************************************************************************/
PCacheFile::PCacheFile(uint32 size) : PFile()
{
	cacheSize     = size;
	cacheFilled   = 0;
	cachePosition = 0;
	cacheStart    = 0;

	cache = new uint8[size];
	if (cache == NULL)
		throw PMemoryException();
}



/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "fileName" is the name of the file to open.                        */
/*         "openFlags" is the open flags you want the file to be opened with. */
/*         "size" is the size of the cache in bytes.                          */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PCacheFile::PCacheFile(PString fileName, uint16 openFlags, uint32 size) : PFile(fileName, openFlags)
{
	cacheSize     = size;
	cacheFilled   = 0;
	cachePosition = 0;
	cacheStart    = 0;

	cache = new uint8[size];
	if (cache == NULL)
		throw PMemoryException();
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PCacheFile::~PCacheFile(void)
{
	delete[] cache;
}



/******************************************************************************/
/* Close() closes the file again.                                             */
/******************************************************************************/
void PCacheFile::Close(void)
{
	// Don't use the cache anymore
	cacheFilled   = 0;
	cachePosition = 0;
	cacheStart    = 0;

	// Call base class
	PFile::Close();
}



/******************************************************************************/
/* Read() will read from the file and store the data into the buffer given.   */
/*                                                                            */
/* Input:  "buffer" is a pointer to the buffer to store the data in.          */
/*         "count" is the number of bytes you want to read.                   */
/*                                                                            */
/* Output: Is the actual number of bytes read.                                */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PCacheFile::Read(void *buffer, int32 count)
{
	int32 readBytes = 0;
	int32 left, todo;

	while (count > 0)
	{
		// Calculate how many bytes left in the cache
		left = cacheFilled - cachePosition;

		// Find the number of bytes we can copy
		todo = min(left, count);

		if (todo != 0)
		{
			// Copy the bytes from the cache we can
			memcpy((uint8 *)buffer + readBytes, cache + cachePosition, todo);

			readBytes     += todo;
			cachePosition += todo;
			count         -= todo;
		}
		else
		{
			// Fill the cache up again
			cacheStart   += cacheFilled;
			cacheFilled   = PFile::Read(cache, cacheSize);
			cachePosition = 0;

			// Check to see if we reached EOF
			if (cacheFilled == 0)
				break;
		}
	}

	return (readBytes);
}



/******************************************************************************/
/* Write() will write into the file from the buffer given.                    */
/*                                                                            */
/* Input:  "buffer" is a pointer to the buffer where the data to write is.    */
/*         "count" is the number of bytes you want to write.                  */
/*                                                                            */
/* Output: Is the actual number of bytes written.                             */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PCacheFile::Write(const void *buffer, int32 count)
{
	// Flush the cache
	cacheFilled   = 0;
	cachePosition = 0;
	cacheStart    = 0;

	return (PFile::Write(buffer, count));
}



/******************************************************************************/
/* Seek() will seek to a new position in the file.                            */
/*                                                                            */
/* Input:  "offset" is the number of bytes you want to move.                  */
/*         "from" is where you want to move from.                             */
/*                                                                            */
/* Output: Is the new position.                                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PCacheFile::Seek(int32 offset, PSeekFlags from)
{
	int32 newPos;

	// Calculate the new real position
	switch (from)
	{
		case pSeekBegin:
			newPos = offset;
			break;

		case pSeekEnd:
			newPos = GetLength() + offset;
			break;

		case pSeekCurrent:
			newPos = cacheStart + cachePosition + offset;
			break;

		default:
			// Unknown seek flag
			ASSERT(false);
			newPos = offset;
			break;
	}

	// Check to see if we need to flush the cache
	if ((newPos < cacheStart) || (newPos >= (cacheStart + cacheFilled)))
	{
		// Flush the cache
		cacheFilled   = 0;
		cachePosition = 0;
		cacheStart    = newPos;

		// Seek to the new position
		PFile::Seek(newPos, pSeekBegin);
	}
	else
	{
		// Calculate new cache position
		cachePosition = newPos - cacheStart;
	}

	// Clear eof flag
	eof = false;

	return (newPos);
}



/******************************************************************************/
/* GetPosition() will return the current file position.                       */
/*                                                                            */
/* Output: Is the current file position.                                      */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PCacheFile::GetPosition(void) const
{
	if (cacheFilled == 0)
		return (PFile::GetPosition());

	return (cacheStart + cachePosition);
}





/******************************************************************************/
/* PMemFile class                                                             */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "growSize" is the number of bytes to grow the buffer with.         */
/******************************************************************************/
PMemFile::PMemFile(uint32 growSize) : PFile()
{
	name       = "Memory File";
	memBuffer  = NULL;
	bufferSize = 0;
	fileSize   = 0;
	position   = 0;
	grow       = growSize;
	fileOpened = true;
}



/******************************************************************************/
/* Constructor                                                                */
/*                                                                            */
/* Input:  "buffer" is a pointer to a buffer to use as a file.                */
/*         "size" is the size of the buffer in bytes.                         */
/*         "growSize" is the number of bytes to grow the buffer with.         */
/******************************************************************************/
PMemFile::PMemFile(uint8 *buffer, uint32 size, uint32 growSize) : PFile()
{
	name      = "Memory File";
	memBuffer = NULL;

	Attach(buffer, size, growSize);
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PMemFile::~PMemFile(void)
{
	Close();
}



/******************************************************************************/
/* Attach() attachs a buffer to use as a file buffer. The buffer must have    */
/*      been allocated with the new uint8[] operator.                         */
/*                                                                            */
/* Input:  "buffer" is a pointer to a buffer to use as a file.                */
/*         "size" is the size of the buffer in bytes.                         */
/*         "growSize" is the number of bytes to grow the buffer with.         */
/******************************************************************************/
void PMemFile::Attach(uint8 *buffer, uint32 size, uint32 growSize)
{
	ASSERT(memBuffer == NULL);

	memBuffer  = buffer;
	bufferSize = size;
	fileSize   = size;
	position   = 0;
	grow       = growSize;

	fileOpened = true;
	eof        = false;
}



/******************************************************************************/
/* Detach() detachs the buffer and return a pointer to it. You is resposible  */
/*      delete the buffer with the delete[] operator.                         */
/*                                                                            */
/* Output: A pointer to the buffer.                                           */
/******************************************************************************/
uint8 *PMemFile::Detach(void)
{
	uint8 *buffer;

	ASSERT(memBuffer != NULL);

	// Remember to pointer
	buffer = memBuffer;

	// Detach the buffer from the class
	memBuffer  = NULL;
	bufferSize = 0;
	fileSize   = 0;
	position   = 0;
	grow       = 0;

	return (buffer);
}



/******************************************************************************/
/* Close() will "close" the file and free the memory buffer.                  */
/******************************************************************************/
void PMemFile::Close(void)
{
	delete[] memBuffer;

	memBuffer  = NULL;
	bufferSize = 0;
	fileSize   = 0;
	position   = 0;
	grow       = 0;
}



/******************************************************************************/
/* Read() will read from the file and store the data into the buffer given.   */
/*                                                                            */
/* Input:  "buffer" is a pointer to the buffer to store the data in.          */
/*         "count" is the number of bytes you want to read.                   */
/*                                                                            */
/* Output: Is the actual number of bytes read.                                */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PMemFile::Read(void *buffer, int32 count)
{
	int32 readBytes;

	ASSERT(fileOpened == true);

	// Calculate how many bytes to read
	readBytes = min(fileSize - position, count);
	if (readBytes <= 0)
	{
		eof = true;
		return (0);
	}

	// Copy the data into the buffer
	memcpy(buffer, memBuffer + position, readBytes);

	// Move the file pointer
	position += readBytes;
	eof = false;

	return (readBytes);
}



/******************************************************************************/
/* Write() will write into the file from the buffer given.                    */
/*                                                                            */
/* Input:  "buffer" is a pointer to the buffer where the data to write is.    */
/*         "count" is the number of bytes you want to write.                  */
/*                                                                            */
/* Output: Is the actual number of bytes written.                             */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PMemFile::Write(const void *buffer, int32 count)
{
	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + count) > bufferSize)
		GrowBuffer(position + count);

	if (count != 0)
	{
		// Write the buffer into the file
		memcpy(memBuffer + position, buffer, count);

		// Calculate new position and file size
		position += count;
		fileSize  = max(fileSize, position);
	}

	return (count);
}



/******************************************************************************/
/* Read_UINT8() reads one byte and return it.                                 */
/*                                                                            */
/* Output: The read byte.                                                     */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint8 PMemFile::Read_UINT8(void)
{
	ASSERT(fileOpened == true);

	// Check for end of file is reached
	if (position >= fileSize)
	{
		eof = true;
		return (0);
	}

	eof = false;
	return (memBuffer[position++]);
}



/******************************************************************************/
/* Read_L_UINT16() reads 2 bytes in little endian and return it in host       */
/*      format.                                                               */
/*                                                                            */
/* Output: The read 16 bit number.                                            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint16 PMemFile::Read_L_UINT16(void)
{
	uint16 retVal;

	ASSERT(fileOpened == true);

	// Check for end of file is reached
	if ((position + 1) >= fileSize)
	{
		eof = true;
		return (0);
	}

	retVal    = P_LENDIAN_TO_HOST_INT16(*((uint16 *)(memBuffer + position)));
	position += 2;
	eof       = false;

	return (retVal);
}



/******************************************************************************/
/* Read_L_UINT32() reads 4 bytes in little endian and return it in host       */
/*      format.                                                               */
/*                                                                            */
/* Output: The read 32 bit number.                                            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint32 PMemFile::Read_L_UINT32(void)
{
	uint32 retVal;

	ASSERT(fileOpened == true);

	// Check for end of file is reached
	if ((position + 3) >= fileSize)
	{
		eof = true;
		return (0);
	}

	retVal    = P_LENDIAN_TO_HOST_INT32(*((uint32 *)(memBuffer + position)));
	position += 4;
	eof       = false;

	return (retVal);
}



/******************************************************************************/
/* ReadArray_L_UINT16s() reads an array of uint16 in little endian and        */
/*      convert them to host format.                                          */
/*                                                                            */
/* Input:  "buffer" is a pointer to where to store the data.                  */
/*         "number" is the number of elements to read.                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::ReadArray_L_UINT16s(uint16 *buffer, int32 number)
{
	uint16 *tempBuffer;

	ASSERT(fileOpened == true);

	// Initialize temp variables
	tempBuffer = (uint16 *)(memBuffer + position);

	for (; ((number > 0) && ((position + 1) < fileSize)); number--, position += 2)
		*buffer++ = P_LENDIAN_TO_HOST_INT16(*tempBuffer++);
}



/******************************************************************************/
/* ReadArray_L_UINT32s() reads an array of uint32 in little endian and        */
/*      convert them to host format.                                          */
/*                                                                            */
/* Input:  "buffer" is a pointer to where to store the data.                  */
/*         "number" is the number of elements to read.                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::ReadArray_L_UINT32s(uint32 *buffer, int32 number)
{
	uint32 *tempBuffer;

	ASSERT(fileOpened == true);

	// Initialize temp variables
	tempBuffer = (uint32 *)(memBuffer + position);

	for (; ((number > 0) && ((position + 3) < fileSize)); number--, position += 4)
		*buffer++ = P_LENDIAN_TO_HOST_INT32(*tempBuffer++);
}



/******************************************************************************/
/* Read_B_UINT16() reads 2 bytes in big endian and return it in host format.  */
/*                                                                            */
/* Output: The read 16 bit number.                                            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint16 PMemFile::Read_B_UINT16(void)
{
	uint16 retVal;

	ASSERT(fileOpened == true);

	// Check for end of file is reached
	if ((position + 1) >= fileSize)
	{
		eof = true;
		return (0);
	}

	retVal    = P_BENDIAN_TO_HOST_INT16(*((uint16 *)(memBuffer + position)));
	position += 2;
	eof       = false;

	return (retVal);
}



/******************************************************************************/
/* Read_B_UINT32() reads 4 bytes in big endian and return it in host format.  */
/*                                                                            */
/* Output: The read 32 bit number.                                            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
uint32 PMemFile::Read_B_UINT32(void)
{
	uint32 retVal;

	ASSERT(fileOpened == true);

	// Check for end of file is reached
	if ((position + 3) >= fileSize)
	{
		eof = true;
		return (0);
	}

	retVal    = P_BENDIAN_TO_HOST_INT32(*((uint32 *)(memBuffer + position)));
	position += 4;
	eof       = false;

	return (retVal);
}



/******************************************************************************/
/* ReadArray_B_UINT16s() reads an array of uint16 in big endian and convert   */
/*      them to host format.                                                  */
/*                                                                            */
/* Input:  "buffer" is a pointer to where to store the data.                  */
/*         "number" is the number of elements to read.                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::ReadArray_B_UINT16s(uint16 *buffer, int32 number)
{
	uint16 *tempBuffer;

	ASSERT(fileOpened == true);

	// Initialize temp variables
	tempBuffer = (uint16 *)(memBuffer + position);

	for (; ((number > 0) && ((position + 1) < fileSize)); number--, position += 2)
	{
		*buffer++ = P_BENDIAN_TO_HOST_INT16(*tempBuffer++);
		position += 2;
	}
}



/******************************************************************************/
/* ReadArray_B_UINT32s() reads an array of uint32 in big endian and convert   */
/*      them to host format.                                                  */
/*                                                                            */
/* Input:  "buffer" is a pointer to where to store the data.                  */
/*         "number" is the number of elements to read.                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::ReadArray_B_UINT32s(uint32 *buffer, int32 number)
{
	uint32 *tempBuffer;

	ASSERT(fileOpened == true);

	// Initialize temp variables
	tempBuffer = (uint32 *)(memBuffer + position);

	for (; ((number > 0) && ((position + 3) < fileSize)); number--, position += 4)
		*buffer++ = P_BENDIAN_TO_HOST_INT32(*tempBuffer++);
}



/******************************************************************************/
/* Write_UINT8() writes one byte.                                             */
/*                                                                            */
/* Input:  "num" is the byte to write.                                        */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::Write_UINT8(uint8 num)
{
	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + 1) > bufferSize)
		GrowBuffer(position + 1);

	memBuffer[position++] = num;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* Write_L_UINT16() writes 2 bytes in little endian.                          */
/*                                                                            */
/* Input:  "num" is the 16 bit number to write.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::Write_L_UINT16(uint16 num)
{
	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + 2) > bufferSize)
		GrowBuffer(position + 2);

	*((uint16 *)(memBuffer + position)) = P_HOST_TO_LENDIAN_INT16(num);
	position += 2;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* Write_L_UINT32() writes 4 bytes in little endian.                          */
/*                                                                            */
/* Input:  "num" is the 32 bit number to write.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::Write_L_UINT32(uint32 num)
{
	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + 4) > bufferSize)
		GrowBuffer(position + 4);

	*((uint32 *)(memBuffer + position)) = P_HOST_TO_LENDIAN_INT32(num);
	position += 4;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* WriteArray_L_UINT16s() writes an array of uint16 in little endian.         */
/*                                                                            */
/* Input:  "buffer" is a pointer to the data to write.                        */
/*         "number" is the number of elements to write.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::WriteArray_L_UINT16s(const uint16 *buffer, int32 number)
{
	uint16 *tempBuf;

	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + number * 2) > bufferSize)
		GrowBuffer(position + number * 2);

	tempBuf = (uint16 *)(memBuffer + position);

	for (int32 t = 0; t < number; t++)
		*(tempBuf + t) = P_HOST_TO_LENDIAN_INT16(*buffer++);

	position += number * 2;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* WriteArray_L_UINT32s() writes an array of uint32 in little endian.         */
/*                                                                            */
/* Input:  "buffer" is a pointer to the data to write.                        */
/*         "number" is the number of elements to write.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::WriteArray_L_UINT32s(const uint32 *buffer, int32 number)
{
	uint32 *tempBuf;

	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + number * 4) > bufferSize)
		GrowBuffer(position + number * 4);

	tempBuf = (uint32 *)(memBuffer + position);

	for (int32 t = 0; t < number; t++)
		*(tempBuf + t) = P_HOST_TO_LENDIAN_INT32(*buffer++);

	position += number * 4;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* Write_B_UINT16() writes 2 bytes in big endian.                             */
/*                                                                            */
/* Input:  "num" is the 16 bit number to write.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::Write_B_UINT16(uint16 num)
{
	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + 2) > bufferSize)
		GrowBuffer(position + 2);

	*((uint16 *)(memBuffer + position)) = P_HOST_TO_BENDIAN_INT16(num);
	position += 2;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* Write_B_UINT32() writes 4 bytes in big endian.                             */
/*                                                                            */
/* Input:  "num" is the 32 bit number to write.                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::Write_B_UINT32(uint32 num)
{
	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + 4) > bufferSize)
		GrowBuffer(position + 4);

	*((uint32 *)(memBuffer + position)) = P_HOST_TO_BENDIAN_INT32(num);
	position += 4;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* WriteArray_B_UINT16s() writes an array of uint16 in big endian.            */
/*                                                                            */
/* Input:  "buffer" is a pointer to the data to write.                        */
/*         "number" is the number of elements to write.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::WriteArray_B_UINT16s(const uint16 *buffer, int32 number)
{
	uint16 *tempBuf;

	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + number * 2) > bufferSize)
		GrowBuffer(position + number * 2);

	tempBuf = (uint16 *)(memBuffer + position);

	for (int32 t = 0; t < number; t++)
		*(tempBuf + t) = P_HOST_TO_BENDIAN_INT16(*buffer++);

	position += number * 2;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* WriteArray_B_UINT32s() writes an array of uint32 in big endian.            */
/*                                                                            */
/* Input:  "buffer" is a pointer to the data to write.                        */
/*         "number" is the number of elements to write.                       */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::WriteArray_B_UINT32s(const uint32 *buffer, int32 number)
{
	uint32 *tempBuf;

	ASSERT(fileOpened == true);

	// Check to see if we need to grow the buffer
	if ((position + number * 4) > bufferSize)
		GrowBuffer(position + number * 4);

	tempBuf = (uint32 *)(memBuffer + position);

	for (int32 t = 0; t < number; t++)
		*(tempBuf + t) = P_HOST_TO_BENDIAN_INT32(*buffer++);

	position += number * 4;
	fileSize = max(fileSize, position);
}



/******************************************************************************/
/* Seek() will seek to a new position in the file.                            */
/*                                                                            */
/* Input:  "offset" is the number of bytes you want to move.                  */
/*         "from" is where you want to move from.                             */
/*                                                                            */
/* Output: Is the new position.                                               */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PMemFile::Seek(int32 offset, PSeekFlags from)
{
	int32 newPos;

	ASSERT(fileOpened == true);

	// Find out where to seek
	switch (from)
	{
		case pSeekBegin:
			newPos = offset;
			break;

		case pSeekCurrent:
			newPos = position + offset;
			break;

		case pSeekEnd:
			newPos = fileSize + offset;
			break;

		default:
			// Unknown seek flag
			ASSERT(false);
			newPos = offset;
			break;
	}

	// Check for a negative position
	if (newPos < 0)
		throw PFileException(P_FILE_ERR_SEEK, name);

	// Use the new position
	position = newPos;

	// Clear eof flag
	eof = false;

	return (newPos);
}



/******************************************************************************/
/* GetPosition() will return the current file position.                       */
/*                                                                            */
/* Output: Is the current file position.                                      */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PMemFile::GetPosition(void) const
{
	ASSERT(fileOpened == true);

	return (position);
}



/******************************************************************************/
/* GetLength() will return the length of the file.                            */
/*                                                                            */
/* Output: Is the length of the file.                                         */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
int32 PMemFile::GetLength(void) const
{
	ASSERT(fileOpened == true);

	return (fileSize);
}



/******************************************************************************/
/* SetLength() will change the length of the file.                            */
/*                                                                            */
/* Input:  "newLength" is the new length of the file.                         */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::SetLength(int32 newLength)
{
	ASSERT(fileOpened == true);

	if (newLength > bufferSize)
		GrowBuffer(newLength);

	fileSize = newLength;
}



/******************************************************************************/
/* GetFileName() will return the filename of the current file.                */
/*                                                                            */
/* Output: Is the filename without the path.                                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PMemFile::GetFileName(void) const
{
	return (name);
}



/******************************************************************************/
/* GetFilePath() will return the expanded path without the filename of the    */
/*      current file. See the GetFullPath() for more information.             */
/*                                                                            */
/* Output: Is the path without the filename.                                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PMemFile::GetFilePath(void) const
{
	return (name);
}



/******************************************************************************/
/* GetFullPath() will return the path with the filename of the current file.  */
/*      It will expand the path if the file is opened with a relative path,   */
/*      e.g. if the current directory is: /boot/home and the relative path is */
/*      ../MyDir/MyDir2/../MyFile.txt   ->   /boot/MyDir/MyFile.txt           */
/*                                                                            */
/* Output: Is the full path with the filename.                                */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PMemFile::GetFullPath(void) const
{
	return (name);
}



/******************************************************************************/
/* GetFileType() will return the type of the file.                            */
/*      BeOS   : The mimestring.                                              */
/*      Windows: The file extension.                                          */
/*      Linux  : ???                                                          */
/*                                                                            */
/* Output: Is the file type.                                                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
PString PMemFile::GetFileType(void) const
{
	// You can't get the filetype on a memory file
	ASSERT(false);

	return ("");
}



/******************************************************************************/
/* SetFilePath() will set the class to use the new filename. It does not open */
/*      the file, it just associates the object with the name.                */
/*                                                                            */
/* Input:  "newName" is the new name to use.                                  */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::SetFilePath(PString newName)
{
	// You can't change the path on a memory file
	ASSERT(false);
}



/******************************************************************************/
/* SetFileType() will change the type of the file.                            */
/*      BeOS   : Change the mimestring.                                       */
/*      Windows: Rename the file?                                             */
/*      Linux  : ???                                                          */
/*                                                                            */
/* Input:  "newType" is a string with the new file type.                      */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::SetFileType(PString newType)
{
	// You can't set the filetype on a memory file
	ASSERT(false);
}



/******************************************************************************/
/* GrowBuffer() will grow the buffer.                                         */
/*                                                                            */
/* Input:  "endPoint" is the end position it will minimum grow to.            */
/*                                                                            */
/* Except: PFileException.                                                    */
/******************************************************************************/
void PMemFile::GrowBuffer(uint32 endPoint)
{
	uint8 *newBuffer;
	uint32 newSize;

	// Check to see if we can grow the buffer
	if (grow == 0)
		throw PFileException(P_FILE_ERR_DISK_FULL, name);

	// Calculate the new size of the buffer
	newSize = bufferSize + grow;
	while (newSize < endPoint)
		newSize += grow;

	// Allocate new buffer
	newBuffer = new uint8[newSize];

	if (fileSize != 0)
	{
		// Copy the old buffer into the new one
		memcpy(newBuffer, memBuffer, fileSize);
	}

	// Free the old buffer and use the new one
	delete[] memBuffer;

	memBuffer  = newBuffer;
	bufferSize = newSize;
}
