/******************************************************************************/
/* PFile header file.                                                         */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#ifndef __PFile_h
#define __PFile_h

// PolyKit headers
#include "POS.h"
#include "PString.h"
#include "ImportExport.h"


/******************************************************************************/
/* PFile class                                                                */
/******************************************************************************/
#if __p_os == __p_beos && __POWERPC__
#pragma export on
#endif

class _IMPEXP_PKLIB PFile
{
public:
	enum POpenFlags
	{
		pModeRead =			0x0000,
		pModeWrite =		0x0001,
		pModeReadWrite =	0x0002,
		pModeCreate =		0x0100,
		pModeNoTruncate =	0x0200,
		pModeFailIfExists =	0x0400
	};

	enum PSeekFlags {pSeekBegin = 0, pSeekCurrent, pSeekEnd};

	PFile(void);
	PFile(PString fileName, uint16 openFlags);
	virtual ~PFile(void);

	virtual void Open(uint16 openFlags);
	virtual void Open(PString fileName, uint16 openFlags);
	virtual void Close(void);

	virtual bool IsOpen(void) const;

	virtual int32 Read(void *buffer, int32 count);
	virtual int32 Write(const void *buffer, int32 count);

	virtual void ReadString(char *buffer, int32 number);
	virtual PString ReadString(void);
	virtual uint8 Read_UINT8(void);

	virtual uint16 Read_L_UINT16(void);
	virtual uint32 Read_L_UINT32(void);
	virtual void ReadArray_L_UINT16s(uint16 *buffer, int32 number);
	virtual void ReadArray_L_UINT32s(uint32 *buffer, int32 number);

	virtual uint16 Read_B_UINT16(void);
	virtual uint32 Read_B_UINT32(void);
	virtual void ReadArray_B_UINT16s(uint16 *buffer, int32 number);
	virtual void ReadArray_B_UINT32s(uint32 *buffer, int32 number);

	virtual void WriteString(PString string);
	virtual void Write_UINT8(uint8 num);

	virtual void Write_L_UINT16(uint16 num);
	virtual void Write_L_UINT32(uint32 num);
	virtual void WriteArray_L_UINT16s(const uint16 *buffer, int32 number);
	virtual void WriteArray_L_UINT32s(const uint32 *buffer, int32 number);

	virtual void Write_B_UINT16(uint16 num);
	virtual void Write_B_UINT32(uint32 num);
	virtual void WriteArray_B_UINT16s(const uint16 *buffer, int32 number);
	virtual void WriteArray_B_UINT32s(const uint32 *buffer, int32 number);

	virtual int32 Seek(int32 offset, PSeekFlags from);
	void SeekToBegin(void);
	void SeekToEnd(void);
	virtual int32 GetPosition(void) const;
	virtual bool IsEOF(void) const;

	virtual int32 GetLength(void) const;
	virtual void SetLength(int32 newLength);

	virtual PString GetFileName(void) const;
	virtual PString GetFilePath(void) const;
	virtual PString GetFullPath(void) const;
	virtual PString GetFileType(void) const;

	virtual void SetFilePath(PString newName);
	virtual void SetFileType(PString newType);

	static void Rename(PString oldName, PString newName);
	static void Remove(PString fileName);
	static bool FileExists(PString fileName);

	virtual PString ReadLine(void);
	virtual void WriteLine(PString line);

protected:
	bool fileOpened;
	bool eof;
	PString name;

#if __p_os == __p_beos

	BFile *file;

#elif __p_os == __p_windows

	HANDLE file;
	uint16 prevOpenFlags;
	int32 prevPos;

#endif
};



/******************************************************************************/
/* PCacheFile class                                                           */
/******************************************************************************/
class _IMPEXP_PKLIB PCacheFile : public PFile
{
public:
	PCacheFile(uint32 size = 32768);
	PCacheFile(PString fileName, uint16 openFlags, uint32 size = 32768);
	virtual ~PCacheFile(void);

	virtual void Close(void);

	virtual int32 Read(void *buffer, int32 count);
	virtual int32 Write(const void *buffer, int32 count);

	virtual int32 Seek(int32 offset, PSeekFlags from);
	virtual int32 GetPosition(void) const;

protected:
	uint8 *cache;
	int32 cacheSize;
	int32 cacheFilled;
	int32 cachePosition;
	int32 cacheStart;
};



/******************************************************************************/
/* PMemFile class                                                             */
/******************************************************************************/
class _IMPEXP_PKLIB PMemFile : public PFile
{
public:
	PMemFile(uint32 growSize = 1024);
	PMemFile(uint8 *buffer, uint32 size, uint32 growSize = 0);
	virtual ~PMemFile(void);

	void Attach(uint8 *buffer, uint32 size, uint32 growSize = 0);
	uint8 *Detach(void);

	virtual void Close(void);

	virtual int32 Read(void *buffer, int32 count);
	virtual int32 Write(const void *buffer, int32 count);

	virtual uint8 Read_UINT8(void);

	virtual uint16 Read_L_UINT16(void);
	virtual uint32 Read_L_UINT32(void);
	virtual void ReadArray_L_UINT16s(uint16 *buffer, int32 number);
	virtual void ReadArray_L_UINT32s(uint32 *buffer, int32 number);

	virtual uint16 Read_B_UINT16(void);
	virtual uint32 Read_B_UINT32(void);
	virtual void ReadArray_B_UINT16s(uint16 *buffer, int32 number);
	virtual void ReadArray_B_UINT32s(uint32 *buffer, int32 number);

	virtual void Write_UINT8(uint8 num);

	virtual void Write_L_UINT16(uint16 num);
	virtual void Write_L_UINT32(uint32 num);
	virtual void WriteArray_L_UINT16s(const uint16 *buffer, int32 number);
	virtual void WriteArray_L_UINT32s(const uint32 *buffer, int32 number);

	virtual void Write_B_UINT16(uint16 num);
	virtual void Write_B_UINT32(uint32 num);
	virtual void WriteArray_B_UINT16s(const uint16 *buffer, int32 number);
	virtual void WriteArray_B_UINT32s(const uint32 *buffer, int32 number);

	virtual int32 Seek(int32 offset, PSeekFlags from);
	virtual int32 GetPosition(void) const;

	virtual int32 GetLength(void) const;
	virtual void SetLength(int32 newLength);

	virtual PString GetFileName(void) const;
	virtual PString GetFilePath(void) const;
	virtual PString GetFullPath(void) const;
	virtual PString GetFileType(void) const;

	virtual void SetFilePath(PString newName);
	virtual void SetFileType(PString newType);

protected:
	void GrowBuffer(uint32 endPoint);

	uint8 *memBuffer;
	int32 bufferSize;
	int32 fileSize;
	int32 position;
	int32 grow;
};

#if __p_os == __p_beos && __POWERPC__
#pragma export off
#endif

#endif
