/******************************************************************************/
/* PThread implementation file.                                               */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#define _BUILDING_POLYKIT_LIBRARY_

// PolyKit headers
#include "PDebug.h"
#include "PException.h"
#include "PThread.h"
#include "GlobalFuncs.h"


/******************************************************************************/
/* PThread class                                                              */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/******************************************************************************/
PThread::PThread(void)
{
	// Initialize member variables
	threadRunning = false;
	threadPri     = pNormal;
	threadName    = "Some PThread";
	threadFunc    = NULL;

#if __p_os == __p_beos

	// Create suspend signal
	suspendSignal = new PEvent(true, true);
	if (suspendSignal == NULL)
		throw PMemoryException();

	suspendCounter = 0;

#elif __p_os == __p_windows

	thread = NULL;

#endif
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PThread::~PThread(void)
{
	ASSERT(threadRunning == false);
}



/******************************************************************************/
/* StartThread() will start the thread.                                       */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
void PThread::StartThread(void)
{
	ASSERT(threadRunning == false);
	ASSERT(threadFunc != NULL);

#if __p_os == __p_beos

	int32 pri;

	// Convert the priority
	pri = ConvertPriority();

	// Spawn the thread
	tid = spawn_thread(ThreadFunc, threadName, pri, this);
	if (tid < 0)
		throw PSystemException(ConvertOSError(tid));

	// Start the thread
	resume_thread(tid);

#elif __p_os == __p_windows

	// Create and start the thread
	if (!(thread = CreateThread(NULL, 0, ThreadFunc, this, 0, &tid)))
		throw PSystemException(ConvertOSError(GetLastError()));

	// Set the thread priority
	if (!SetThreadPriority(thread, ConvertPriority()))
		throw PSystemException(ConvertOSError(GetLastError()));

	// Wait until the thread have started running
	threadStart.Lock();

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* WaitOnThread() will wait until the thread exits.                           */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
void PThread::WaitOnThread(void) const
{
	// Exit if the thread is not running
	if (!threadRunning)
		return;

#if __p_os == __p_beos

	status_t exitVal, retVal;

	// First wait on the suspend signal. If it is set, the thread is running
	if (suspendSignal->Lock() == pSyncError)
		throw PSystemException(P_ERR_ANY);

	// Wait on the thread
	retVal = wait_for_thread(tid, &exitVal);
	if (retVal != B_OK)
		throw PSystemException(ConvertOSError(retVal));

#elif __p_os == __p_windows

	DWORD res;

	// Wait on the thread
	res = WaitForSingleObject(thread, INFINITE);
	if (res != WAIT_OBJECT_0)
	{	
		printf("Err = %d\n", GetLastError());

		// Failed for some reason
		if (res == WAIT_FAILED)
			throw PSystemException(ConvertOSError(GetLastError()));

		ASSERT(false);
	}

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* GetExitCode() will return the code your thread returned.                   */
/*                                                                            */
/* Output: The exit code.                                                     */
/******************************************************************************/
int32 PThread::GetExitCode(void) const
{
	ASSERT(threadRunning == false);

	return (exitCode);
}



/******************************************************************************/
/* Resume() will resume the thread if it was suspended.                       */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
void PThread::Resume(void)
{
	ASSERT(threadRunning == true);

#if __p_os == __p_beos

	status_t retVal;

	// Decrement the suspend counter
	if (AtomicDecrement(&suspendCounter) == 0)
	{
		// Set the suspend signal
		if (suspendSignal->SetEvent() == pSyncError)
			throw PSystemException(P_ERR_ANY);

		// Start the thread
		retVal = resume_thread(tid);
		if (retVal != B_OK)
			throw PSystemException(ConvertOSError(retVal));
	}

#elif __p_os == __p_windows

	// Resume the thread
	if (ResumeThread(thread) == 0xFFFFFFFF)
		throw PSystemException(ConvertOSError(GetLastError()));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Suspend() will suspend the thread.                                         */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
void PThread::Suspend(void)
{
	ASSERT(threadRunning == true);

#if __p_os == __p_beos

	status_t retVal;

	// Increment the suspend counter
	if (AtomicIncrement(&suspendCounter) == 1)
	{
		// Reset the suspend signal
		if (suspendSignal->ResetEvent() == pSyncError)
			throw PSystemException(P_ERR_ANY);

		// Suspend the thread
		retVal = suspend_thread(tid);
		if (retVal != B_OK)
			throw PSystemException(ConvertOSError(retVal));
	}

#elif __p_os == __p_windows

	// Suspend the thread
	if (SuspendThread(thread) == 0xFFFFFFFF)
		throw PSystemException(ConvertOSError(GetLastError()));

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* Kill() will try to kill the thread.                                        */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
void PThread::Kill(void)
{
	ASSERT(threadRunning == true);

#if __p_os == __p_beos

	status_t retVal;

	// Kill the thread
	retVal = kill_thread(tid);
	if (retVal != B_OK)
		throw PSystemException(ConvertOSError(retVal));

	// Clear the suspend counter
	suspendCounter = 0;
	suspendSignal->SetEvent();

#elif __p_os == __p_windows

	BOOL ok;

	// Terminate the thread
	ok = TerminateThread(thread, exitCode);

	// Close the handle
	VERIFY(CloseHandle(thread));

	// Did we succeed?
	if (!ok)
		throw PSystemException(ConvertOSError(GetLastError()));

#elif __p_os == __p_linux
implement
#endif

	// The thread isn't running anymore
	threadRunning = false;
}



/******************************************************************************/
/* IsAlive() tells if the thread is still running.                            */
/******************************************************************************/
bool PThread::IsAlive(void) const
{
	return (threadRunning);
}



/******************************************************************************/
/* SetHookFunc() will tell the class what function to run.                    */
/*                                                                            */
/* Input:  "func" is a pointer to the function.                               */
/*         "userData" can be anything you desire. It will be given to your    */
/*         function as argument.                                              */
/******************************************************************************/
void PThread::SetHookFunc(PThreadFunc func, void *userData)
{
	ASSERT(func != NULL);

	// Remember the data
	threadFunc = func;
	threadData = userData;
}



/******************************************************************************/
/* SetName() will change the thread name.                                     */
/*                                                                            */
/* Input:  "name" is the new thread name.                                     */
/******************************************************************************/
void PThread::SetName(PString name)
{
	threadName = name;
}



/******************************************************************************/
/* SetPriority() will change the thread priority. If the thread is running,   */
/*      it will change the priority immediately.                              */
/*                                                                            */
/* Input:  "pri" is the new priority.                                         */
/*                                                                            */
/* Except: PSystemException.                                                  */
/******************************************************************************/
void PThread::SetPriority(PPriority pri)
{
	// Remember the priority
	threadPri = pri;

#if __p_os == __p_beos

	// Set the thread priority if the thread is running
	if (threadRunning)
	{
		status_t retVal;

		retVal = set_thread_priority(tid, ConvertPriority());
		if (retVal != B_OK)
			throw PSystemException(ConvertOSError(retVal));
	}

#elif __p_os == __p_windows

	// Set the thread priority if the thread is running
	if (threadRunning)
	{
		if (!SetThreadPriority(thread, ConvertPriority()))
			throw PSystemException(ConvertOSError(GetLastError()));
	}

#elif __p_os == __p_linux
implement
#endif
}



/******************************************************************************/
/* GetPriority() will return the current priority of the thread.              */
/*                                                                            */
/* Output: Is the current priority.                                           */
/******************************************************************************/
PThread::PPriority PThread::GetPriority(void) const
{
	return (threadPri);
}



/******************************************************************************/
/* ThreadFunc() is the thread function that will be started. It will call the */
/*      user function and remember the exit code.                             */
/*                                                                            */
/* Input:  "object" is a pointer to the current PThread object.               */
/******************************************************************************/
#if __p_os == __p_beos

int32 PThread::ThreadFunc(void *object)
{
	PThread *obj = (PThread *)object;

	// Call the user thread
	obj->threadRunning = true;
	obj->exitCode      = obj->threadFunc(obj->threadData);
	obj->threadRunning = false;

	return (obj->exitCode);
}

#elif __p_os == __p_windows

DWORD WINAPI PThread::ThreadFunc(void *object)
{
	PThread *obj = (PThread *)object;

	// Call the user thread
	obj->threadRunning = true;
	obj->threadStart.SetEvent();	// Set the thread start event
	obj->exitCode      = obj->threadFunc(obj->threadData);
	obj->threadRunning = false;

	// Close the handle
	VERIFY(CloseHandle(obj->thread));

	return (obj->exitCode);
}

#endif



/******************************************************************************/
/* ConvertPriority() will convert a PolyKit priority to an OS priority.       */
/*                                                                            */
/* Output: Is the OS priority.                                                */
/******************************************************************************/
#if __p_os == __p_beos

int32 PThread::ConvertPriority(void) const
{
	int32 pri;

	switch (threadPri)
	{
		case pIdle:
			pri = 1;
			break;

		case pLow:
			pri = B_LOW_PRIORITY;
			break;

		case pBelowNormal:
			pri = B_NORMAL_PRIORITY - ((B_NORMAL_PRIORITY - B_LOW_PRIORITY) / 2);
			break;

		case pAboveNormal:
			pri = B_NORMAL_PRIORITY + ((B_NORMAL_PRIORITY - B_LOW_PRIORITY) / 2);
			break;

		case pHigh:
			pri = B_DISPLAY_PRIORITY;
			break;

		case pRealTime:
			pri = B_REAL_TIME_PRIORITY;
			break;

		default:
			pri = B_NORMAL_PRIORITY;
			break;
	}

	return (pri);
}

#elif __p_os == __p_windows

int PThread::ConvertPriority(void) const
{
	int pri;

	switch (threadPri)
	{
		case pIdle:
			pri = THREAD_PRIORITY_IDLE;
			break;

		case pLow:
			pri = THREAD_PRIORITY_LOWEST;
			break;

		case pBelowNormal:
			pri = THREAD_PRIORITY_BELOW_NORMAL;
			break;

		case pAboveNormal:
			pri = THREAD_PRIORITY_ABOVE_NORMAL;
			break;

		case pHigh:
			pri = THREAD_PRIORITY_HIGHEST;
			break;

		case pRealTime:
			pri = THREAD_PRIORITY_TIME_CRITICAL;
			break;

		default:
			pri = THREAD_PRIORITY_NORMAL;
			break;
	}

	return (pri);
}

#endif
