/******************************************************************************/
/* PTupleSpace implementation file.                                           */
/******************************************************************************/
/* This is part of the PolyKit library.                                       */
/* Copyright (C) 1998-1999 by Polycode.                                       */
/* All rights reserved.                                                       */
/*                                                                            */
/* This source, or parts thereof, may be used for free in free- and shareware */
/* programs. For commercial programs, you have to pay a fee to use it. See    */
/* the license file for more information.                                     */
/******************************************************************************/


#define _BUILDING_POLYKIT_LIBRARY_

// PolyKit header
#include "PException.h"
#include "PTupleSpace.h"


/******************************************************************************/
/* PTupleSpace class                                                          */
/******************************************************************************/

/******************************************************************************/
/* Constructor                                                                */
/******************************************************************************/
PTupleSpace::PTupleSpace(void)
{
	// Allocate a skip list for the item entries
	entryList = new PSkipList<PList<PEntry> *, PString>(5);
	if (entryList == NULL)
		throw PMemoryException();
}



/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/
PTupleSpace::~PTupleSpace(void)
{
	// Empty the tuple space
	MakeEmpty();

	// Delete the skip list containing the item entries
	delete entryList;
}



/******************************************************************************/
/* AddBool() adds a bool to the tuple space.                                  */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aBool" is the bool to add to the tuple entry.                     */
/******************************************************************************/
void PTupleSpace::AddBool(PString name, bool aBool)
{
	PItem item;
	item.aBool = aBool;
	AddItem(name, pTypeBool, item);
}



/******************************************************************************/
/* AddInt8() adds an int8 to the tuple space.                                 */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aInt8" is the int8 to add to the tuple entry.                     */
/******************************************************************************/
void PTupleSpace::AddInt8(PString name, int8 aInt8)
{
	PItem item;
	item.aInt8 = aInt8;
	AddItem(name, pTypeInt8, item);
}



/******************************************************************************/
/* AddInt16() adds an int16 to the tuple space.                               */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aInt16" is the int16 to add to the tuple entry.                   */
/******************************************************************************/
void PTupleSpace::AddInt16(PString name, int16 aInt16)
{
	PItem item;
	item.aInt16 = aInt16;
	AddItem(name, pTypeInt16, item);
}



/******************************************************************************/
/* AddInt32() adds an int32 to the tuple space.                               */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aInt16" is the int32 to add to the tuple entry.                   */
/******************************************************************************/
void PTupleSpace::AddInt32(PString name, int32 aInt32)
{
	PItem item;
	item.aInt32 = aInt32;
	AddItem(name, pTypeInt32, item);
}



/******************************************************************************/
/* AddInt64() adds an int64 to the tuple space.                               */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aInt64" is the int64 to add to the tuple entry.                   */
/******************************************************************************/
void PTupleSpace::AddInt64(PString name, int64 aInt64)
{
	PItem item;
	item.aInt64 = aInt64;
	AddItem(name, pTypeInt64, item);
}



/******************************************************************************/
/* AddFloat() adds a float to the tuple space.                                */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aFloat" is the float to add to the tuple entry.                   */
/******************************************************************************/
void PTupleSpace::AddFloat(PString name, float aFloat)
{
	PItem item;
	item.aFloat = aFloat;
	AddItem(name, pTypeFloat, item);
}



/******************************************************************************/
/* AddDouble() adds a double to the tuple space.                              */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aDouble" is the double to add to the tuple entry.                 */
/******************************************************************************/
void PTupleSpace::AddDouble(PString name, double aDouble)
{
	PItem item;
	item.aDouble = aDouble;
	AddItem(name, pTypeDouble, item);
}



/******************************************************************************/
/* AddPointer() adds a pointer to the tuple space.                            */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aPointer" is the pointer to add to the tuple entry.               */
/******************************************************************************/
void PTupleSpace::AddPointer(PString name, const void *aPointer)
{
	PItem item;
	item.pointer = (void *)aPointer;
	AddItem(name, pTypePointer, item);
}



/******************************************************************************/
/* RemoveBool() removes a bool from the tuple space.                          */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the bool type in the tuple entry.         */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemoveBool(PString name, int32 index)
{
	RemoveItem(name, pTypeBool, index);
}



/******************************************************************************/
/* RemoveInt8() removes an int8 from the tuple space.                         */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int8 type in the tuple entry.         */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemoveInt8(PString name, int32 index)
{
	RemoveItem(name, pTypeInt8, index);
}



/******************************************************************************/
/* RemoveInt16() removes an int16 from the tuple space.                       */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int16 type in the tuple entry.        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemoveInt16(PString name, int32 index)
{
	RemoveItem(name, pTypeInt16, index);
}



/******************************************************************************/
/* RemoveInt32() removes an int32 from the tuple space.                       */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int32 type in the tuple entry.        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemoveInt32(PString name, int32 index)
{
	RemoveItem(name, pTypeInt32, index);
}



/******************************************************************************/
/* RemoveInt64() removes an int64 from the tuple space.                       */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int64 type in the tuple entry.        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemoveInt64(PString name, int32 index)
{
	RemoveItem(name, pTypeInt64, index);
}



/******************************************************************************/
/* RemoveFloat() removes a float from the tuple space.                        */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the float type in the tuple entry.        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemoveFloat(PString name, int32 index)
{
	RemoveItem(name, pTypeFloat, index);
}



/******************************************************************************/
/* RemoveDouble() removes a double from the tuple space.                      */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the double type in the tuple entry.       */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemoveDouble(PString name, int32 index)
{
	RemoveItem(name, pTypeDouble, index);
}



/******************************************************************************/
/* RemovePointer() removes a pointer from the tuple space.                    */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the pointer type in the tuple entry.      */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemovePointer(PString name, int32 index)
{
	RemoveItem(name, pTypePointer, index);
}



/******************************************************************************/
/* RemoveName() removes a tuple entry together with all it's items.           */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*                                                                            */
/* Except: PKeyException.                                                     */
/******************************************************************************/
void PTupleSpace::RemoveName(PString name)
{
	delete entryList->GetAndRemoveItem(name);
}



/******************************************************************************/
/* SetBool() sets a new bool at a tuple entry.                                */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "newBool" is the new bool to set to the tuple entry.               */
/*         "index" is the index for the bool type in the tuple entry.         */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetBool(PString name, bool newBool, int32 index)
{
	PItem item;
	item.aBool = newBool;
	SetItem(name, pTypeBool, item, index);
}



/******************************************************************************/
/* SetInt8() sets a new int8 at a tuple entry.                                */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "newInt8" is the new int8 to set to the tuple entry.               */
/*         "index" is the index for the int8 type in the tuple entry.         */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetInt8(PString name, int8 newInt8, int32 index)
{
	PItem item;
	item.aInt8 = newInt8;
	SetItem(name, pTypeInt8, item, index);
}



/******************************************************************************/
/* SetInt16() sets a new int16 at a tuple entry.                              */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "newInt16" is the new int16 to set to the tuple entry.             */
/*         "index" is the index for the int16 type in the tuple entry.        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetInt16(PString name, int16 newInt16, int32 index)
{
	PItem item;
	item.aInt16 = newInt16;
	SetItem(name, pTypeInt16, item, index);
}



/******************************************************************************/
/* SetInt32() sets a new int32 at a tuple entry.                              */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "newInt32" is the int32 to set to the tuple entry.                 */
/*         "index" is the index for the int32 type in the tuple entry.        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetInt32(PString name, int32 newInt32, int32 index)
{
	PItem item;
	item.aInt32 = newInt32;
	SetItem(name, pTypeInt32, item, index);
}



/******************************************************************************/
/* SetInt64() sets a new int64 at a tuple entry.                              */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "newInt64" is the new int64 to set to the tuple entry.             */
/*         "index" is the index for the int64 type in the tuple entry.        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetInt64(PString name, int64 newInt64, int32 index)
{
	PItem item;
	item.aInt64 = newInt64;
	SetItem(name, pTypeInt64, item, index);
}



/******************************************************************************/
/* SetFloat() sets a new float at a tuple entry.                              */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "newFloat" is the float to set to the tuple entry.                 */
/*         "index" is the index for the float type in the tuple entry.        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetFloat(PString name, float newFloat, int32 index)
{
	PItem item;
	item.aFloat = newFloat;
	SetItem(name, pTypeFloat, item, index);
}



/******************************************************************************/
/* SetDouble() sets a new double at a tuple entry.                            */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aDouble" is the double to set to the tuple entry.                 */
/*         "index" is the index for the double type in the tuple entry.       */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetDouble(PString name, double newDouble, int32 index)
{
	PItem item;
	item.aDouble = newDouble;
	SetItem(name, pTypeDouble, item, index);
}



/******************************************************************************/
/* SetPointer() sets a new pointer at a tuple entry.                          */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "aPointer" is the pointer to set to the tuple entry.               */
/*         "index" is the index for the pointer type in the tuple entry.      */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetPointer(PString name, const void *newPointer, int32 index)
{
	PItem item;
	item.pointer = (void *)newPointer;
	SetItem(name, pTypePointer, item, index);
}



/******************************************************************************/
/* GetBool() returns a bool from a tuple entry.                               */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the bool type in the tuple entry.         */
/*                                                                            */
/* Output: the bool to receive.                                               */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
bool PTupleSpace::GetBool(PString name, int32 index)
{
	return (GetItem(name, pTypeBool, index).aBool);
}



/******************************************************************************/
/* GetInt8() returns an int8 from a tuple entry.                              */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int8 type in the tuple entry.         */
/*                                                                            */
/* Output: the int8 to receive.                                               */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int8 PTupleSpace::GetInt8(PString name, int32 index)
{
	return (GetItem(name, pTypeInt8, index).aInt8);
}



/******************************************************************************/
/* GetInt16() returns an int16 from a tuple entry.                            */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int16 type in the tuple entry.        */
/*                                                                            */
/* Output: the int16 to receive.                                              */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int16 PTupleSpace::GetInt16(PString name, int32 index)
{
	return (GetItem(name, pTypeInt16, index).aInt16);
}



/******************************************************************************/
/* GetInt32() returns an int32 from a tuple entry.                            */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int32 type in the tuple entry.        */
/*                                                                            */
/* Output: the int32 to receive.                                              */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int32 PTupleSpace::GetInt32(PString name, int32 index)
{
	return (GetItem(name, pTypeInt32, index).aInt32);
}



/******************************************************************************/
/* GetInt64() returns an int64 from a tuple entry.                            */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int64 type in the tuple entry.        */
/*                                                                            */
/* Output: the int64 to receive.                                              */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int64 PTupleSpace::GetInt64(PString name, int32 index)
{
	return (GetItem(name, pTypeInt64, index).aInt64);
}



/******************************************************************************/
/* GetFloat() returns an float from a tuple entry.                            */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the float type in the tuple entry.        */
/*                                                                            */
/* Output: the float to receive.                                              */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
float PTupleSpace::GetFloat(PString name, int32 index)
{
	return (GetItem(name, pTypeFloat, index).aFloat);
}



/******************************************************************************/
/* GetDouble() returns an double from a tuple entry.                          */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the double type in the tuple entry.       */
/*                                                                            */
/* Output: the double to receive.                                             */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
double PTupleSpace::GetDouble(PString name, int32 index)
{
	return (GetItem(name, pTypeDouble, index).aDouble);
}



/******************************************************************************/
/* GetPointer() returns an pointer from a tuple entry.                        */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the pointer type in the tuple entry.      */
/*                                                                            */
/* Output: the pointer to receive.                                            */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
const void* PTupleSpace::GetPointer(PString name, int32 index)
{
	return (GetItem(name, pTypePointer, index).pointer);
}



/******************************************************************************/
/* GetAndSetBool() returns the old bool from a tuple entry and sets the tuple */
/*      entry to a new bool.                                                  */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the bool type in the tuple entry.         */
/*         "newBool" is the new bool to set at the tuple entry.               */
/*                                                                            */
/* Output: the old bool to receive.                                           */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
bool PTupleSpace::GetAndSetBool(PString name, bool newBool, int32 index)
{
	PItem item;
	item.aBool = newBool;
	item = GetAndSetItem(name, pTypeBool, item, index);
	return (item.aBool);
}



/******************************************************************************/
/* GetAndSetInt8() returns the old int8 from a tuple entry and sets the tuple */
/*      entry to a new int8.                                                  */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int8 type in the tuple entry.         */
/*         "newInt8" is the new int8 to set at the tuple entry.               */
/*                                                                            */
/* Output: the old int8 to receive.                                           */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int8 PTupleSpace::GetAndSetInt8(PString name, int8 newInt8, int32 index)
{
	PItem item;
	item.aInt8 = newInt8;
	item = GetAndSetItem(name, pTypeInt8, item, index);
	return (item.aInt8);
}



/******************************************************************************/
/* GetAndSetInt16() returns the old int16 from a tuple entry and sets the     */
/*      tuple entry to a new int16.                                           */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int16 type in the tuple entry.        */
/*         "newInt16" is the new int16 to set at the tuple entry.             */
/*                                                                            */
/* Output: the old int16 to receive.                                          */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int16 PTupleSpace::GetAndSetInt16(PString name, int16 newInt16, int32 index)
{
	PItem item;
	item.aInt16 = newInt16;
	item = GetAndSetItem(name, pTypeInt16, item, index);
	return (item.aInt16);
}



/******************************************************************************/
/* GetAndSetInt32() returns the old int32 from a tuple entry and sets the     */
/*      tuple entry to a new int32.                                           */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int32 type in the tuple entry.        */
/*         "newInt32" is the new int32 to set at the tuple entry.             */
/*                                                                            */
/* Output: the old int32 to receive.                                          */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int32 PTupleSpace::GetAndSetInt32(PString name, int32 newInt32, int32 index)
{
	PItem item;
	item.aInt32 = newInt32;
	item = GetAndSetItem(name, pTypeInt32, item, index);
	return (item.aInt32);
}



/******************************************************************************/
/* GetAndSetInt64() returns the old int64 from a tuple entry and sets the     */
/*      tuple entry to a new int64.                                           */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int64 type in the tuple entry.        */
/*         "newInt64" is the new int64 to set at the tuple entry.             */
/*                                                                            */
/* Output: the old int64 to receive.                                          */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int64 PTupleSpace::GetAndSetInt64(PString name, int64 newInt64, int32 index)
{
	PItem item;
	item.aInt64 = newInt64;
	item = GetAndSetItem(name, pTypeInt64, item, index);
	return (item.aInt64);
}



/******************************************************************************/
/* GetAndSetFloat() returns the old float from a tuple entry and sets the     */
/*      tuple entry to a new float.                                           */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the float type in the tuple entry.        */
/*         "newFloat" is the new float to set at the tuple entry.             */
/*                                                                            */
/* Output: the old float to receive.                                          */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
float PTupleSpace::GetAndSetFloat(PString name, float newFloat, int32 index)
{
	PItem item;
	item.aFloat = newFloat;
	item = GetAndSetItem(name, pTypeFloat, item, index);
	return (item.aFloat);
}



/******************************************************************************/
/* GetAndSetDouble() returns the old double from a tuple entry and sets the   */
/*      tuple entry to a new double.                                          */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the double type in the tuple entry.       */
/*         "newDouble" is the new double to set at the tuple entry.           */
/*                                                                            */
/* Output: the old double to receive.                                         */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
double PTupleSpace::GetAndSetDouble(PString name, double newDouble, int32 index)
{
	PItem item;
	item.aDouble = newDouble;
	item = GetAndSetItem(name, pTypeDouble, item, index);
	return (item.aDouble);
}



/******************************************************************************/
/* GetAndSetPointer() returns the old pointer from a tuple entry and sets the */
/*      tuple entry to a new pointer.                                         */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the pointer type in the tuple entry.      */
/*         "newPointer" is the new pointer to set at the tuple entry.         */
/*                                                                            */
/* Output: the old pointer to receive.                                        */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
const void* PTupleSpace::GetAndSetPointer(PString name, const void *newPointer, int32 index)
{
	PItem item;
	item.pointer = (void *)newPointer;
	item = GetAndSetItem(name, pTypePointer, item, index);
	return (item.pointer);
}


	
/******************************************************************************/
/* GetAndRemoveBool() returns a bool from a tuple entry and removes it from   */
/*      the tuple entry.                                                      */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the bool type in the tuple entry.         */
/*                                                                            */
/* Output: the bool to receive.                                               */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
bool PTupleSpace::GetAndRemoveBool(PString name, int32 index)
{
	return (GetAndRemoveItem(name, pTypeBool, index).aBool);
}



/******************************************************************************/
/* GetAndRemoveInt8() returns a int8 from a tuple entry and removes it from   */
/*      the tuple entry.                                                      */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int8 type in the tuple entry.         */
/*                                                                            */
/* Output: the int8 to receive.                                               */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int8 PTupleSpace::GetAndRemoveInt8(PString name, int32 index)
{
	return (GetAndRemoveItem(name, pTypeInt8, index).aInt8);
}



/******************************************************************************/
/* GetAndRemoveInt16() returns a int16 from a tuple entry and removes it from */
/*      the tuple entry.                                                      */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int16 type in the tuple entry.        */
/*                                                                            */
/* Output: the int16 to receive.                                              */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int16 PTupleSpace::GetAndRemoveInt16(PString name, int32 index)
{
	return (GetAndRemoveItem(name, pTypeInt16, index).aInt16);
}



/******************************************************************************/
/* GetAndRemoveInt32() returns a int32 from a tuple entry and removes it from */
/*      the tuple entry.                                                      */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int32 type in the tuple entry.        */
/*                                                                            */
/* Output: the int32 to receive.                                              */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int32 PTupleSpace::GetAndRemoveInt32(PString name, int32 index)
{
	return (GetAndRemoveItem(name, pTypeInt32, index).aInt32);
}



/******************************************************************************/
/* GetAndRemoveInt64() returns a int64 from a tuple entry and removes it from */
/*      the tuple entry.                                                      */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the int64 type in the tuple entry.        */
/*                                                                            */
/* Output: the int64 to receive.                                              */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
int64 PTupleSpace::GetAndRemoveInt64(PString name, int32 index)
{
	return (GetAndRemoveItem(name, pTypeInt64, index).aInt64);
}



/******************************************************************************/
/* GetAndRemoveFloat() returns a float from a tuple entry and removes it from */
/*      the tuple entry.                                                      */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the float type in the tuple entry.        */
/*                                                                            */
/* Output: the float to receive.                                              */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
float PTupleSpace::GetAndRemoveFloat(PString name, int32 index)
{
	return (GetAndRemoveItem(name, pTypeFloat, index).aFloat);
}



/******************************************************************************/
/* GetAndRemoveDouble() returns a double from a tuple entry and removes it    */
/*      from the tuple entry.                                                 */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the double type in the tuple entry.       */
/*                                                                            */
/* Output: the double to receive.                                             */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
double PTupleSpace::GetAndRemoveDouble(PString name, int32 index)
{
	return (GetAndRemoveItem(name, pTypeDouble, index).aDouble);
}



/******************************************************************************/
/* GetAndRemovePointer() returns a pointer from a tuple entry and removes it  */
/*      from the tuple entry.                                                 */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "index" is the index for the pointer type in the tuple entry.      */
/*                                                                            */
/* Output: the pointer to receive.                                            */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
const void* PTupleSpace::GetAndRemovePointer(PString name, int32 index)
{
	return (GetAndRemoveItem(name, pTypePointer, index).pointer);
}



/******************************************************************************/
/* CountNames() counts and returns the number of names in the tuple space.    */
/*      from the tuple entry.                                                 */
/*                                                                            */
/* Output: the number of names in the tuple space.                            */
/******************************************************************************/
uint32 PTupleSpace::CountNames(void) const
{
	return (entryList->CountItems());
}



/******************************************************************************/
/* GetNames() gets the names for the item types contained in the tuple space  */
/*                                                                            */
/* Input:  "list" is a pointer to a list of strings where the names will be   */
/*             stored into.                                                   */
/*                                                                            */
/* Output: the number of names in the tuple space.                            */
/******************************************************************************/
/*
void PTupleSpace::GetNames(PList<PString> *list)
{
	entryList->GetKeys(list);
}
*/



/******************************************************************************/
/* MakeEmpty() empties the tuple space.                                       */
/******************************************************************************/
void PTupleSpace::MakeEmpty(void)
{
	PList<PString> keyList;

	// Get the keys of the entry list
	entryList->GetKeys(&keyList);

	// Delete all the items in the entry list using the keys
	for (int32 i = 0; i < keyList.CountItems(); i++)
		delete entryList->GetAndRemoveItem(keyList.GetItem(i));
}



/******************************************************************************/
/* IsEmpty() tests if the tuple space is empty.                               */
/*                                                                            */
/* Output: true if the tuple space is empty, or false otherwise.              */
/******************************************************************************/
bool PTupleSpace::IsEmpty(void)
{
	return (entryList->IsEmpty());
}



/******************************************************************************/
/* AddItem() adds an item to the tuple space.                                 */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "type" is the type of the item.                                    */
/*         "item" is the item to add to the tuple entry.                      */
/******************************************************************************/
void PTupleSpace::AddItem(PString name, PType type, PItem item)
{
	PEntry entry;
	PList<PEntry> *list;

	// Try to get the tuple space entry with the specified name
	try
	{
		// Return the list containing the items
		list = entryList->GetItem(name);

		// Initialize a new entry
		entry.type = type;
		entry.item = item;

		// Add the new entry to the list
		list->AddTail(entry);
	}
	// If a key exception occur, then the entry have not been inserted yet
	catch (PKeyException e)
	{
		// The assumption is that the entry could not be found
		ASSERT(e.errorNum == P_KEY_ERR_ENTRY_NOT_FOUND);

		// Allocate a new list for the item entries
		list = new PList<PEntry>;
		if (list == NULL)
			throw PMemoryException();	// Throw memory exception on failure

		// Insert the new list into the entry list
		entryList->InsertItem(name, list);

		// Make a new attempt to add the item
		AddItem(name, type, item);
	}
}



/******************************************************************************/
/* RemoveItem() removes an item from the tuple space at a specific index.     */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "type" is the type of the item.                                    */
/*         "index" is the index for item type in the tuple entry.             */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::RemoveItem(PString name, PType type, int32 index)
{
	PList<PEntry> *list;
	PEntry entry;
	int32 count, numItems;

	// Return the list containing the item
	list = entryList->GetItem(name);

	// Count the number of items in the list
	numItems = list->CountItems();

	// Throw exception if the index is out of bounds
	if ((index < 0) || (index > numItems))
		throw PBoundsException(P_ERR_ANY, index);

	// Run through all the items in the list
	count = 0;	// Reset counter
	for (int32 i = 0; i < numItems; i++)
	{
		// Get the next entry in the list
		entry = list->GetItem(i);

		// If the entry type is of the specified item type, then..
		if (entry.type == type)
		{
			// If the counter is equal to the specified index, then..
			if (count == index)
			{
				// Remove the item from the list at the specified index
				list->RemoveItem(i);

				// If the previous number of items in the list was 1,
				// the delete the list of entries in the entry list
				if (numItems == 1)
					delete entryList->GetAndRemoveItem(name);

				// Exit the function
				return;
			}
			// Else increment the counter
			else
				count++;
		}
	}

	// Throw out of bounds exception as the index was not found
	throw PBoundsException(P_ERR_ANY, index);
}



/******************************************************************************/
/* SetItem() sets annew item in the tuple space at a specific index.          */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "type" is the type of the item.                                    */
/*         "newItem" is the new item to set at the tuple entry.               */
/*         "index" is the index for item type in the tuple entry.             */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
void PTupleSpace::SetItem(PString name, PType type, PItem newItem, int32 index)
{
	PList<PEntry> *list;
	PEntry entry;
	int32 count, numItems;

	// Return the list containing the item
	list = entryList->GetItem(name);

	// Count the number of items in the list
	numItems = list->CountItems();

	// Throw exception if the index is out of bounds
	if ((index < 0) || (index > numItems))
		throw PBoundsException(P_ERR_ANY, index);

	// Run through all the items in the list
	count = 0;	// Reset counter
	for (int32 i = 0; i < numItems; i++)
	{
		// Get the next entry in the list
		entry = list->GetItem(i);

		// If the entry type is of the specified item type, then..
		if (entry.type == type)
		{
			// If the counter is equal to the specified index, then..
			if (count == index)
			{
				// Set the item of the entry to the new item
				entry.item = newItem;
				list->SetItem(entry, i);

				// Exit the function
				return;
			}
			// Else increment the counter
			else
				count++;
		}
	}

	// Throw out of bounds exception as the index was not found
	throw PBoundsException(P_ERR_ANY, index);
}



/******************************************************************************/
/* GetItem() returns an item from the tuple space at a specific index.        */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "type" is the type of the item.                                    */
/*         "index" is the index for item type in the tuple entry.             */
/*                                                                            */
/* Output: the item to receive.                                               */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
PTupleSpace::PItem PTupleSpace::GetItem(PString name, PType type, int32 index)
{
	PList<PEntry> *list;
	PEntry entry;
	int32 count, numItems;

	// Return the list containing the item
	list = entryList->GetItem(name);

	// Count the number of items in the list
	numItems = list->CountItems();

	// Throw exception if the index is out of bounds
	if ((index < 0) || (index > numItems))
		throw PBoundsException(P_ERR_ANY, index);

	// Run through all the items in the list
	count = 0;	// Reset counter
	for (int32 i = 0; i < list->CountItems(); i++)
	{
		// Get the next entry in the list
		entry = list->GetItem(i);

		// If the entry type is of the specified item type, then..
		if (entry.type == type)
		{
			// If the counter is equal to the specified index, then return the
			// item at the entry 
			if (count == index)
				return (entry.item);

			// Else increment the counter
			else
				count++;
		}
	}

	// Throw out of bounds exception as the index was not found
	throw PBoundsException(P_ERR_ANY, index);
}



/******************************************************************************/
/* GetAndSetItem() returns an old item from the tuple space and sets a new    */
/*      item at the tuple entry.                                              */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "type" is the type of the item.                                    */
/*         "newItem" is the new item to set at the tuple entry.               */
/*         "index" is the index for item type in the tuple entry.             */
/*                                                                            */
/* Output: the item to receive.                                               */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
PTupleSpace::PItem PTupleSpace::GetAndSetItem(PString name, PType type, PItem newItem, int32 index)
{
	PList<PEntry> *list;
	PEntry entry;
	int32 count, numItems;

	// Return the list containing the item
	list = entryList->GetItem(name);

	// Count the number of items in the list
	numItems = list->CountItems();

	// Throw exception if the index is out of bounds
	if ((index < 0) || (index > numItems))
		throw PBoundsException(P_ERR_ANY, index);

	// Run through all the items in the list
	count = 0;	// Reset counter
	for (int32 i = 0; i < list->CountItems(); i++)
	{
		// Get the next entry in the list
		entry = list->GetItem(i);

		// If the entry type is of the specified item type, then..
		if (entry.type == type)
		{
			// If the counter is equal to the specified index, then..
			if (count == index)
			{
				// Store the old item at the entry
				PItem oldItem = entry.item;

				// Set the new item at the entry
				entry.item = newItem;
				list->SetItem(entry, i);

				// Return the old item
				return (oldItem);
			}
			// Else increment the counter
			else
				count++;
		}
	}

	// Throw out of bounds exception as the index was not found
	throw PBoundsException(P_ERR_ANY, index);
}



/******************************************************************************/
/* GetAndRemoveItem() returns an item from the tuple space and removes it     */
/*      from the tuple entry.                                                 */
/*                                                                            */
/* Input:  "name" is the name of the tuple entry.                             */
/*         "type" is the type of the item.                                    */
/*         "index" is the index for item type in the tuple entry.             */
/*                                                                            */
/* Output: the item to receive.                                               */
/*                                                                            */
/* Except: PKeyException, PBoundsException.                                   */
/******************************************************************************/
PTupleSpace::PItem PTupleSpace::GetAndRemoveItem(PString name, PType type, int32 index)
{
	PList<PEntry> *list;
	PEntry entry;
	int32 count, numItems;

	// Return the list containing the item
	list = entryList->GetItem(name);

	// Count the number of items in the list
	numItems = list->CountItems();

	// Throw exception if the index is out of bounds
	if ((index < 0) || (index > numItems))
		throw PBoundsException(P_ERR_ANY, index);

	// Run through all the items in the list
	count = 0;	// Reset counter
	for (int32 i = 0; i < numItems; i++)
	{
		// Get the next entry in the list
		entry = list->GetItem(i);

		// If the entry type is of the specified item type, then..
		if (entry.type == type)
		{
			// If the counter is equal to the specified index, then..
			if (count == index)
			{
				// Store the old item at the entry
				PItem oldItem = entry.item;

				// Remove the item from the list of entries
				list->RemoveItem(i);

				// If the previous number of items in the list was 1,
				// the delete the list of entries in the entry list
				if (numItems == 1)
					delete entryList->GetAndRemoveItem(name);

				// Return the old item
				return (oldItem);
			}
			// Else increment the counter
			else
				count++;
		}
	}

	// Throw out of bounds exception as the index was not found
	throw PBoundsException(P_ERR_ANY, index);
}
